import numpy as np
import itertools
import datetime
import sys
from scipy.stats import t
from scipy.stats import f

from omars_analysis.subset_selection import get_soe
from omars_analysis.generate_model_matrix_heredity import create_model_matrix_heredity

def hat_matrix(smat):
    hat = smat @ np.linalg.pinv(smat.T @ smat) @ smat.T
    return hat

def create_quadratic_interactions(smat):
    no_runs, nf = smat.shape[0], smat.shape[1]
    quad_ix = [i for i,x in enumerate(smat.T) if (abs(x)!=1).any()]
    
    # Quadratic
    if len(quad_ix) > 0:
        quadratic = np.absolute(smat[:,quad_ix])
        soe_mat = np.hstack((np.ones((no_runs,1)),quadratic))
    else:
        soe_mat = np.ones((no_runs,1))

    # Interactions
    all_interactions = []
    for combi in itertools.combinations(range(nf), 2):
        col = (smat[:,combi[0]]*smat[:,combi[1]]).reshape((no_runs,1))
        soe_mat = np.hstack((soe_mat, col))
        all_interactions.append(str(combi[0]+1)+'_'+str(combi[1]+1))

    #centering (important for later functions)
    soe_mat = soe_mat - soe_mat.mean(axis=0)

    return soe_mat[:,1:], all_interactions, quad_ix

def code(smat):
    a = np.ones((smat.shape[0],1))
    for i in smat.T:
        maxi = max(i)
        mini = min(i)
        avg = (maxi + mini)/2
        diff = (maxi - mini)/2
        temp = (i - avg)/diff
        temp = temp.reshape((smat.shape[0],1))
        a = np.hstack((a,temp))
       
    return a[:,1:]

def get_omars_analysis(smat, cy, qheredity: str, iheredity: str, effects_to_drop: list, full='y'):

    # mat (numpy array) - design matrix
    # cy (numpy array) - response (single response, the response will be centered in the function)
    # qheredity (string) - 
    # iheredity (string) - 
    # full (string) - ('y', 'n'); n - only main effects selection
    # effects_to_drop (list) - list of second order effects to be excluded, eg: ['1_1', '2_3']

    mat = code(smat)
    no_runs, nf = mat.shape[0], mat.shape[1]

    # Centering response in case not centered already
    cy = (cy - cy.mean(axis=0)).reshape(no_runs,1)

    # Calculate error variance
    intercept = np.ones((no_runs, 1))
    mat_qi, all_interactions, quad_ix = create_quadratic_interactions(mat)
    total_mat = np.hstack((np.hstack((intercept, mat)),mat_qi))
    
    new_error_variance_projection = np.identity(no_runs) - (total_mat @ np.linalg.pinv(total_mat))
    denominator = no_runs - np.linalg.matrix_rank(total_mat)
    print('\nInitial error degrees of freedom - {}'.format(denominator))

    new_variance = (cy.T @ new_error_variance_projection @ cy)/denominator
    print('\nInitial estimate of error variance - {}'.format(float(new_variance)))
    
    s = np.sqrt(new_variance)
    
    '''
    ANALYSIS OF ME (STEP 1)
    '''
    mebeta = (np.linalg.inv(mat.T @ mat) @ mat.T @ cy).reshape(1,nf)
    vdiags = np.matrix.diagonal(np.linalg.inv(mat.T @ mat))
    svdiags = np.sqrt(vdiags).reshape(1,nf)
    tvalue = np.absolute(mebeta/(s*svdiags))
    ttest = t.ppf(1 - (0.05/2), df=denominator)
    tcomp = tvalue/ttest
    
    indexrme = [i for i,x in enumerate(tcomp[0]) if (abs(x)>=1)] # index of active main effects    
    indexfme = [i for i,x in enumerate(tcomp[0]) if (abs(x)<1)] # index of inactive main effects
    active = mat[:, indexrme]
    inactive = mat[:, indexfme]
    new_me = [str(g+1) for g in indexrme]
    print('\nActive main effects - {}'.format(new_me))

    if full == 'n':
        return None
    
    '''
    END OF STEP 1
    '''
    # Update error variance
    new_variance_projection_updated = new_error_variance_projection + hat_matrix(inactive)
    numerator = cy.T @ new_variance_projection_updated @ cy
    new_denominator = denominator + len(indexfme)
    new_variance = numerator/new_denominator
    print('\nUpdated estimate of error variance - {}'.format(float(new_variance)))
    
    '''
    ANALYSIS OF SOE (STEP 2)
    '''
    # Second order proection matrix (make sure to only multiply with centered response since intercept is ignored)
    y_second_projection = np.identity(no_runs) - hat_matrix(mat) - new_error_variance_projection
    cy_second = (y_second_projection @ cy).reshape(no_runs,1)
    TSS = cy_second.T @ cy_second # Total sum of squares (for second order space)
    
    ####        F Test for second order effects
    o_dfleft = np.linalg.matrix_rank(mat_qi) # quadratics need to be centered in mat_qi
    ftest = f.ppf(q= 1-0.2, dfn=o_dfleft, dfd=new_denominator)
    F1 = (TSS/o_dfleft)/new_variance
    fcomp = F1/ftest
    '''
    CREATE SECOND ORDER MATRIX WITH HEREDITY FOR ANALYSIS (WITHOUT INTERCEPT)
    '''
    if fcomp >= 1:
        # standardize
        std = 'n'
        soe_mat2, my_dictionary_soe_single  = create_model_matrix_heredity(std, mat, mat_qi, quad_ix, active, inactive, indexrme, indexfme, qheredity, iheredity)
        if len(effects_to_drop) > 0:
            new_dict = {}
            reduced_matrix_indices = []
            ittr = 0
            for keyz in my_dictionary_soe_single.items():
                if keyz[0] in effects_to_drop:
                    continue
                else:
                    reduced_matrix_indices.append(keyz[1])
                    new_dict.update({keyz[0]:ittr})
                    ittr += 1
            soe_mat2 = soe_mat2[:,reduced_matrix_indices]
            my_dictionary_soe_single = new_dict
            
        
        # if quadratic_special_coding == 'y':
        #     relevant_q = []
        #     for dumzy in my_dictionary_soe_single:
        #         if dumzy.split('_')[0] == dumzy.split('_')[1]:
        #             relevant_q.append(int(dumzy.split('_')[0])-1)
                    
        #     if len(relevant_q)>0:
        #         temp_mat = mat[:,relevant_q]
        #         temp_mat = temp_mat - temp_mat.mean(axis=0) # center
        #         temp_mat = np.square(temp_mat) # square
        #         temp_mat = code(temp_mat) # normalize
        #         soe_mat2[:,0:len(relevant_q)] = temp_mat.copy()

        i_start = datetime.datetime.now()
        active_soe, p_value_omars = get_soe(mat, o_dfleft, soe_mat2, cy_second, new_denominator, new_variance, my_dictionary_soe_single)
        i_finish = datetime.datetime.now()
        timing = str((i_finish - i_start).total_seconds())
    else:
        active_soe = []
        p_value_omars = np.nan
    
    new_qe = []
    new_ie = []
    for d in active_soe:
        x = d.split('_')
        if x[0] == x[1]:
            new_qe.append(d)
        else:
            new_ie.append(d)

    print('\nActive interaction effects - {}'.format(new_ie))
    print('\nActive quadratic effects - {}'.format(new_qe))

    return p_value_omars


import numpy as np
import itertools
import scipy
from scipy.stats import f

def get_key(val, x): 
    for key, value in x.items(): 
         if val == value: 
             return key   

    return "key doesn't exist"




def get_soe(mat, limit_for_soe_subset, soe_mat2, cy_second, denominator, new_variance, my_dictionary_soe_single):  
    # List of active SOE terms
    active_soe = []

    dfleft = limit_for_soe_subset

    actual_soe_df = np.linalg.matrix_rank(soe_mat2)
    N_by_4 = int(mat.shape[0]/4)
    if soe_mat2.shape[1] <= limit_for_soe_subset or actual_soe_df <= N_by_4:
        dummy5 = actual_soe_df
    else:
        dummy5 = N_by_4

    # Skipped 'intercept' as response is centered
    
    ###################################################################
    ############    CODE FOR SUBSET SELECTION  #######################
    ###################################################################
    # m = subset size in consideration
    # id = last combination with lowest RSS
    # soe_mat2 does not have the intercept
    # dummy5 = # of df left for second order terms
    m = 0
    for n in range(dummy5):
        m += 1
        first = 0
        for combi in itertools.combinations(range(soe_mat2.shape[1]), m):
            dummy7 = np.ones((soe_mat2.shape[0],1))
            if m == 1:
                dummy8 = soe_mat2[:,combi[0]]
                dummy8.shape = (soe_mat2.shape[0],1)
                dummy7 = np.hstack((dummy7, dummy8))
                v_key = combi[0]
                resid = np.linalg.lstsq(dummy7, cy_second, rcond=None)[1]
                if combi[0] == 0:
                    l_rss = resid
                    l_v_key = v_key                
            else:
                dummy8 = soe_mat2[:,combi]
                dummy7 = np.hstack((dummy7, dummy8))
                resid = np.linalg.lstsq(dummy7, cy_second, rcond=None)[1]
                if resid.size == 0:
                    continue
                v_key = combi
                if first == 0:
                    l_rss = resid
                    l_v_key = v_key
                    first = 1
            if resid < l_rss:
                l_rss = resid
                l_v_key = v_key
        dfleft -= 1
        ftest = f.ppf(q= 1-0.2, dfn=(dfleft), dfd=denominator)
        F2 = (l_rss/(dfleft))/new_variance
        fcomp = F2/ftest
        fcomp = float(fcomp)
        if m == soe_mat2.shape[1] or fcomp < 1:
            break

    p_value_omars = 1-scipy.stats.f.cdf(fcomp, dfleft, denominator)
    
    if m == 1:
        active_soe.append(get_key(l_v_key, my_dictionary_soe_single))
    else:
        for s in range(len(l_v_key)):
            active_soe.append(get_key(l_v_key[s], my_dictionary_soe_single))

    return active_soe, p_value_omars


import numpy as np
import itertools

'''
Inputs
qheredity = y,n
iheredity = s,w,n

'''

def create_model_matrix_heredity(std, mat, mat_qi, quad_ix, active, real_inactive, indexrme, indexfme, qheredity, iheredity):
    # my_dictionary_soe_ci - individual column indices for each soe column
    # my_dictionary_soe_single - location of a certain soe column in the soe matrix
    
    no_runs, nf = mat.shape[0], mat.shape[1]
    
    my_dictionary_soe_ci = {}
    my_dictionary_soe_single = {}
    if active.shape[1] == 0:# or active.shape[1] == 1:
        qheredity = 'n'
        iheredity = 'n'
    
    if qheredity == 'y':
        # print('\nHeredity for quadratic effects assumed')
        k=0
        ittr = 0
        active_q = []
        for j in active.T:
            if np.any(j == 0):
                v_key = str(indexrme[k]+1)+"_"+str(indexrme[k]+1)
                active_q.append(k)
                # current_dictionary = {v_key: [indexrme[k], indexrme[k]]}
                # my_dictionary_soe_ci.update(current_dictionary)
                current_dictionary2 = {v_key: ittr}
                my_dictionary_soe_single.update(current_dictionary2)
                ittr += 1
            k += 1
            
        soe_mat2 = np.absolute(active[:,active_q])
    else:
        # print('\nNo Heredity assumed for quadratic effects')
        no_q = len(quad_ix)
        soe_mat2 = np.absolute(mat_qi[:,:no_q])
        
        for j in range(soe_mat2.shape[1]):
            v_key = str(quad_ix[j]+1)+"_"+str(quad_ix[j]+1)
            current_dictionary = {v_key: [quad_ix[j], quad_ix[j]]}
            my_dictionary_soe_ci.update(current_dictionary)
            current_dictionary2 = {v_key: j}
            my_dictionary_soe_single.update(current_dictionary2)
        
    
    # Pattern - conventional
    interactions = []
    if iheredity == 's':
        # print('\nStrong Heredity assumed for interactions')    
        for combi in itertools.combinations(range(active.shape[1]), 2):
            iron = (active[:,combi[0]] * active[:,combi[1]]).reshape((no_runs,1))
            
            # iron = np.multiply(active[:,combi[0]], active[:,combi[1]])
            # iron = np.reshape(iron, (iron.shape[0],1))       
            soe_mat2 = np.hstack((soe_mat2, iron))
            v_key = str(indexrme[combi[0]]+1)+"_"+str(indexrme[combi[1]]+1)
            interactions.append(v_key)
            # current_dictionary = {v_key: [indexrme[combi[0]], indexrme[combi[1]]]}
            # my_dictionary_soe_ci.update(current_dictionary)
            current_dictionary2 = {v_key: soe_mat2.shape[1]-1}
            my_dictionary_soe_single.update(current_dictionary2)
    # Pattern - Corrected
    elif iheredity == 'w':
        # index = indexrme + indexfme
        # print('\nWeak Heredity assumed for interactions')
        for combi in itertools.combinations(range(active.shape[1]), 2):
            iron = (active[:,combi[0]] * active[:,combi[1]]).reshape((no_runs,1))
            
            # iron = np.multiply(active[:,combi[0]], active[:,combi[1]])
            # iron = np.reshape(iron, (iron.shape[0],1))       
            soe_mat2 = np.hstack((soe_mat2, iron))
            v_key = str(indexrme[combi[0]]+1)+"_"+str(indexrme[combi[1]]+1)
            interactions.append(v_key)
            # current_dictionary = {v_key: [indexrme[combi[0]], indexrme[combi[1]]]}
            # my_dictionary_soe_ci.update(current_dictionary)
            current_dictionary2 = {v_key: soe_mat2.shape[1]-1}
            my_dictionary_soe_single.update(current_dictionary2)
        for combi2 in range(active.shape[1]):
            for combi3 in range(real_inactive.shape[1]):
                o = (active[:, [combi2]] * active[:, [combi3]]).reshape((no_runs,1))
                # master1 = active[:, combi2]
                # master1.shape = (active.shape[0],1)
                # master2 = real_inactive[:, combi3]
                # master2.shape = (active.shape[0],1)
                # o = np.multiply(master1,master2)
                # o.shape = (o.shape[0],1)                                 
                soe_mat2 = np.hstack((soe_mat2, o))
                mini, maxi = min(indexrme[combi2], indexfme[combi3]), max(indexrme[combi2], indexfme[combi3])
                v_key = str(mini+1)+"_"+str(maxi+1)
                interactions.append(v_key)
                # current_dictionary = {v_key: [mini, maxi]}                   
                # my_dictionary_soe_ci.update(current_dictionary)
                current_dictionary2 = {v_key: soe_mat2.shape[1]-1}
                my_dictionary_soe_single.update(current_dictionary2)
    # Pattern - conventional
    else:
        # print('\nNo heredity assumed for interactions')
        for combi in itertools.combinations(range(mat.shape[1]), 2):
            iron = np.multiply(mat[:,combi[0]], mat[:,combi[1]])
            iron = np.reshape(iron, (iron.shape[0],1))       
            soe_mat2 = np.hstack((soe_mat2, iron))
            v_key = str(combi[0]+1)+"_"+str(combi[1]+1)
            interactions.append(v_key)
            # current_dictionary = {v_key: [combi[0], combi[1]]}
            # my_dictionary_soe_ci.update(current_dictionary)
            current_dictionary2 = {v_key: soe_mat2.shape[1]-1}
            my_dictionary_soe_single.update(current_dictionary2)
    
    if std == 'y':
        soe_mat2 = (soe_mat2 - np.mean(soe_mat2, axis=0)) / np.std(soe_mat2, axis=0)
    else:
        soe_mat2 = soe_mat2 - np.mean(soe_mat2, axis=0)
        
    return soe_mat2, my_dictionary_soe_single



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['omars_analysis']

package_data = \
{'': ['*'], 'omars_analysis': ['data/*']}

install_requires = \
['numpy>=1.23.5,<2.0.0', 'scipy>=1.9.3,<2.0.0']

setup_kwargs = {
    'name': 'omars-analysis',
    'version': '0.1.2',
    'description': '',
    'long_description': "# Documentation for package omars-analysis\n\nThis package has one function 'get_omars_analysis'\n\nFirst run the following:\n\n```python\nfrom omars_analysis.main import get_omars_analysis\n```\n\nRunning the above command will make the make the function 'get_omars_analysis' available for use.\n\n## Function usage\n\nThe function requires six inputs:\n\n- mat\n  \n  the first input is the design matrix. The design matrix need not be coded, however it must only consist of continuous variables. The function is meant to be used with designs that have all factors either with two levels or three levels. The design should **NOT** consist of headers. The design matrix should not consist of second order effects since this will be built internally in the function.\n- cy\n  \n  This is the response.\n\n- qheredity\n  \n  This is to specify heredity constraints for quadratic effects. The accepted inputs are 'y' or 'n' ('y'- strong heredity, 'n'- no heredity, 'n'- No heredity). The input must be a string in lowercase.\n- iheredity\n  \n  This is to specify heredity constraints for two-factor interaction effects. The accepted inputs are 's', 'w' or 'n' ('s'- strong heredity, 'w'- weak heredity, 'n'- no heredity). The input must be a string in lowercase.\n\n- effects_to_drop\n  \n  This is to specify second order effects that must be excluded from the analysis. The input must be a list of strings. For example: ['1_1', '2_3']. This input specifies that the quadratic effect of the first factor and the interaction effect between factor two and three must be excluded from the second step of the analysis (subset selection).\n\n- full\n  \n  'n' -  analysis is performed on the main effects only\n  \n  'y' - analysis is performed on the main effects and second order effects.\n\n  The default is set to 'y'\n\n## Output\n\nThe function will auttomatically print out the following:\n\n- Initial error degrees of freedom available\n- The initial estimate of the error variance\n- Main effects that are active\n- Updated estimate of the error variance\n- Active interaction effects\n- Active quadratic effects\n\nThe function outputs one return value. This value is the p-value from the last failed F-test during the second order effects selection.\n\n## Example code\n\n```python\noutput = get_omars_analysis(mat=design_matrix, cy=response, qheredity='n', iheredity='n', effects_to_drop=[], full='y')\n```\n",
    'author': 'Mohammed Saif Ismail Hameed',
    'author_email': 'saifismailh@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
